theory Xmlt2
  imports Extended_Nat Xmlt
begin

(*TODO: move ?*)
lemma sum_mono [partial_function_mono]:
  assumes p1: "\<And>x. mono_option (p1 x)"
      and p2: "\<And>x. mono_option (p2 x)"
      and f: "mono_option f"
  shows "mono_option (\<lambda>g. case s of Inl x \<Rightarrow> p1 x g | Inr x \<Rightarrow> p2 x g)"
    (is "mono_option ?f")
  using assms by (cases s, auto intro!:partial_function_mono)

datatype 'a xml_error = TagMismatch "tag list" | Fatal 'a
text {*
  @{const TagMismatch} represents tag mismatch at root position. It is allowed when
  the tag is optional.
*}

lemma xml_error_mono [partial_function_mono]:
  assumes p1: "\<And>x. mono_option (p1 x)"
      and p2: "\<And>x. mono_option (p2 x)"
      and f: "mono_option f"
  shows "mono_option (\<lambda>g. case s of TagMismatch x \<Rightarrow> p1 x g | Fatal x \<Rightarrow> p2 x g)"
    (is "mono_option ?f")
  using assms by (cases s, auto intro!:partial_function_mono)

type_synonym 'a xmlt2 = "xml \<Rightarrow> (string xml_error + 'a) option"
type_synonym 'a xmlst = "xml list \<Rightarrow> (string xml_error + 'a) option"

definition tag_mismatch where
  "tag_mismatch tag xs \<equiv> Some (Inl (TagMismatch tag))"

definition fatal where
  "fatal str xs \<equiv> Some (Inl (Fatal str))"
hide_const (open) fatal

text {* Turn @{typ "'a xmlt2"} into @{typ "'a xmlt"} by disallowing tag mismatch. *}
definition sum_bot_of_option where [code del]:
 "sum_bot_of_option f opt \<equiv>
  case opt of None \<Rightarrow> Bottom
  | Some x \<Rightarrow> f x"

lemma sum_bot_of_option_code [code]: "sum_bot_of_option f (Some x) = f x"
  by (simp add: sum_bot_of_option_def)

definition xmlt_of_xmlt2 :: "'a xmlt2 \<Rightarrow> 'a xmlt" where
 "xmlt_of_xmlt2 p xml \<equiv>
  sum_bot_of_option (\<lambda>ret.
    case ret
    of Inl (TagMismatch tags) \<Rightarrow> Left (''Tag mismatch: expecting '' @ show tags)
    | Inl (Fatal str) \<Rightarrow> Left str
    | Inr a \<Rightarrow> Right a) (p xml)"


text {* Conditional parsing depending on tag match. *}

definition xml_cond :: "'a xmlt2 \<Rightarrow> 'a xmlt2 \<Rightarrow> 'a xmlt2" (infixr "XMLor" 51)
where
 "xml_cond p1 p2 xml \<equiv>
  p1 xml \<bind> (\<lambda>ret1. case ret1
    of Inl err1 \<Rightarrow> (case err1
      of TagMismatch tags1 \<Rightarrow>
        p2 xml \<bind> (\<lambda>ret2. case ret2
          of Inl err2 \<Rightarrow> Some (Inl (
            case err2 of TagMismatch tags2 \<Rightarrow> TagMismatch (tags1 @ tags2)
            | _ \<Rightarrow> err2))
          | _ \<Rightarrow> Some ret2)
      | _ \<Rightarrow> Some ret1)
    | _ \<Rightarrow> Some ret1)"

lemma xml_cond_mono [partial_function_mono]:
  assumes "\<And>y. mono_option (p1 y)" and "\<And>y. mono_option (p2 y)"
  shows "mono_option (\<lambda>g. xml_cond (\<lambda>y. p1 y g) (\<lambda> y. p2 y g) x)"
  using assms by (auto intro!: partial_function_mono simp: xml_cond_def)

definition xml_begin :: "tag \<Rightarrow> 'a xmlst \<Rightarrow> 'a xmlt2" where
  "xml_begin tag p xml \<equiv>
   case xml of XML nam atts xs \<Rightarrow> if nam = tag then p xs else tag_mismatch [tag] xml
   | _ \<Rightarrow> Some (Inl (Fatal (''expecting <'' @ show tag @ ''> but encountered text element '' @ show xml)))"

lemma xml_begin_mono [partial_function_mono]:
  assumes "\<And>y. mono_option (p y)"
  shows "mono_option (\<lambda>g. xml_begin tag (\<lambda>y. p y g) x)"
proof (cases x)
  case (XML t atts xs)
  with assms show ?thesis by (cases atts, auto intro!: partial_function_mono simp: xml_begin_def)
next
  case (XML_text str)
  then show ?thesis by (auto intro!: partial_function_mono simp: xml_begin_def)
qed

fun xml_return :: "'a \<Rightarrow> 'a xmlst" where
  "xml_return v [] = Some (Inr v)"
| "xml_return v (x#xs) = Some (Inl (Fatal (''want to close but found\<newline>'' @ show x)))"

definition "disallow_tag_mismatch err x \<equiv>
  Inl (Fatal (case err of TagMismatch tags \<Rightarrow>
    ''expecting '' @ show tags @ '' but encountered '' @ show x
  | Fatal x \<Rightarrow> x))"

text {* parses the first mandatory child *}
fun xml_take :: "'a xmlt2 \<Rightarrow> ('a \<Rightarrow> 'b xmlst) \<Rightarrow> 'b xmlst" where
  "xml_take p1 p2 [] = Some (Inl (Fatal (''unexpected tag close'')))"
| "xml_take p1 p2 (x # xs) =
  p1 x \<bind>
  (\<lambda>ret. case ret of
    Inl err \<Rightarrow> Some (disallow_tag_mismatch err x)
  | Inr a \<Rightarrow> p2 a xs )"

lemma xml_take_mono [partial_function_mono]:
  assumes "\<And>y. mono_option (p1 y)"
      and "\<And>a. mono_option (p2 a (tl xs))"
  shows "mono_option (\<lambda>g. xml_take (\<lambda>y. p1 y g) (\<lambda>a ys. p2 a ys g) xs)"
  using assms by (cases xs, auto intro!:partial_function_mono)

text {* Enters into first mandatory child *}
definition xml_enter :: "tag \<Rightarrow> 'a xmlst \<Rightarrow> ('a \<Rightarrow> 'b xmlst) \<Rightarrow> 'b xmlst" where
  "xml_enter tag p1 p2 = xml_take (xml_begin tag p1) p2"

lemma xml_enter_mono [partial_function_mono]:
  assumes "\<And>y. mono_option (p1 y)"
      and "\<And>a. mono_option (p2 a (tl xs))"
  shows "mono_option (\<lambda>g. xml_enter tag (\<lambda>y. p1 y g) (\<lambda>a ys. p2 a ys g) xs)"
  using assms by (auto simp: xml_enter_def intro!:partial_function_mono)


definition xml_take_text :: "(string \<Rightarrow> 'a xmlst) \<Rightarrow> 'a xmlst" where
  "xml_take_text p xs \<equiv>
   case xs of [] \<Rightarrow> Some (Inl (Fatal (''expecting a text but tag closed'')))
   | XML_text text # xmls \<Rightarrow> p text xmls
   | x#xs \<Rightarrow> Some (Inl (Fatal (''expecting a text but encountered\<newline>'' @ show x)))"

lemma xml_take_text_mono [partial_function_mono]:
  assumes "\<And>text. mono_option (p text (tl xs))"
  shows "mono_option (\<lambda>g. xml_take_text (\<lambda>text ys. p test ys g) xs)"
proof (cases xs)
  case Nil
  then show ?thesis by (auto intro!: partial_function_mono simp: xml_take_text_def)
next
  case (Cons x list)
  with assms show ?thesis by (cases x, auto intro!: partial_function_mono simp: xml_take_text_def)
qed

definition xml_take_int :: "(int \<Rightarrow> 'a xmlst) \<Rightarrow> 'a xmlst" where
"xml_take_int p xs \<equiv>
  case xs of [] \<Rightarrow> Some (Inl (Fatal (''expecting an integer but tag closed'')))
  | XML_text text # xs \<Rightarrow>
    (case int_of_string text of Left x \<Rightarrow> Some (Inl (Fatal x)) | Right n \<Rightarrow> p n xs)
  | x#xs \<Rightarrow> Some (Inl (Fatal (''expecting an integer but encountered\<newline>'' @ show x)))"

lemma xml_take_int_mono [partial_function_mono]:
  assumes "\<And>n. mono_option (p n (tl xs))"
  shows "mono_option (\<lambda>g. xml_take_int (\<lambda>n ys. p n ys g) xs)"
proof (cases xs)
  case Nil
  then show ?thesis by (auto intro!: partial_function_mono simp: xml_take_int_def)
next
  case xs: (Cons x list)
  show ?thesis
  proof (cases x)
    case x: (XML x11 x12 x13)
    show ?thesis by (auto intro!: partial_function_mono simp: xml_take_int_def xs x)
  next
    case x: (XML_text str)
    from assms show ?thesis 
      by (cases "int_of_string str", auto intro!: partial_function_mono simp: xml_take_int_def xs x)
  qed
qed

text {*
  Parses the first child, if tag matches.
*}
fun xml_take_optional :: "'a xmlt2 \<Rightarrow> ('a option \<Rightarrow> 'b xmlst) \<Rightarrow> 'b xmlst" where
  "xml_take_optional p1 p2 [] = p2 None []"
| "xml_take_optional p1 p2 (x # xs) =
    p1 x \<bind>
    (\<lambda>ret. case ret of Inl err \<Rightarrow>
       (case err of TagMismatch _ \<Rightarrow> p2 None (x#xs) (* TagMismatch is allowed *)
        | _ \<Rightarrow> Some (Inl err))
     | Inr a \<Rightarrow> p2 (Some a) xs)"

fun xml_take_default :: "'a \<Rightarrow> 'a xmlt2 \<Rightarrow> ('a \<Rightarrow> 'b xmlst) \<Rightarrow> 'b xmlst" where
  "xml_take_default a p1 p2 [] = p2 a []"
| "xml_take_default a p1 p2 (x # xs) =
    p1 x \<bind>
    (\<lambda>ret. case ret of Inl err \<Rightarrow>
       (case err of TagMismatch _ \<Rightarrow> p2 a (x#xs) (* TagMismatch is allowed *)
        | _ \<Rightarrow> Some (Inl err))
     | Inr a \<Rightarrow> p2 a xs)"

lemma xml_take_optional_mono [partial_function_mono]:
  assumes p1: "\<And>y. mono_option (p1 y)"
      and p2: "\<And>a ys. mono_option (p2 a ys)"
  shows "mono_option (\<lambda>g. xml_take_optional (\<lambda>y. p1 y g) (\<lambda>a ys. p2 a ys g) xs)"
proof (cases xs)
  case Nil
  with assms show ?thesis by (auto intro!: partial_function_mono)
next
  case (Cons x list)
  with assms show ?thesis by (cases x, auto intro!: partial_function_mono)
qed

lemma xml_take_default_mono [partial_function_mono]:
  assumes p1: "\<And>y. mono_option (p1 y)"
      and p2: "\<And>a ys. mono_option (p2 a ys)"
  shows "mono_option (\<lambda>g. xml_take_default a (\<lambda>y. p1 y g) (\<lambda>a ys. p2 a ys g) xs)"
proof (cases xs)
  case Nil
  with assms show ?thesis by (auto intro!: partial_function_mono)
next
  case (Cons x list)
  with assms show ?thesis by (cases x, auto intro!: partial_function_mono)
qed

text {* Take first children, as many as tag matches. *}

fun xml_take_many_sub :: "'a list \<Rightarrow> nat \<Rightarrow> enat \<Rightarrow> 'a xmlt2 \<Rightarrow> ('a list \<Rightarrow> 'b xmlst) \<Rightarrow> 'b xmlst" where
  "xml_take_many_sub acc minOccurs maxOccurs p1 p2 [] =
  ( if minOccurs = 0 then p2 (rev acc) []
    else Some (Inl (Fatal (''minOccurs failure''))))"
| "xml_take_many_sub acc minOccurs maxOccurs p1 p2 (x # xs) =
  ( if maxOccurs = 0 then p2 (rev acc) (x#xs)
    else
      p1 x \<bind> (\<lambda>ret.
        case ret of Inl err \<Rightarrow> (
          case err of TagMismatch tags \<Rightarrow>
            if minOccurs = 0 then p2 (rev acc) (x # xs)
            else Some (Inl (Fatal (''minOccurs failure: '' @ show tags)))
          | _ \<Rightarrow> Some (Inl err))
       | Inr a \<Rightarrow> xml_take_many_sub (a # acc) (minOccurs-1) (maxOccurs-1) p1 p2 xs))"

abbreviation xml_take_many where "xml_take_many \<equiv> xml_take_many_sub []"

lemma xml_take_many_mono [partial_function_mono]:
  assumes p1: "\<And>ys. mono_option (p1 ys)"
      and p2: "\<And>a ys. mono_option (p2 a ys)"
  shows "mono_option (\<lambda>g. xml_take_many_sub acc minOccurs maxOccurs (\<lambda>ys. p1 ys g) (\<lambda>a ys. p2 a ys g) xs)"
proof(induct xs arbitrary: acc minOccurs maxOccurs)
  case Nil
  from assms show ?case by (auto intro!: partial_function_mono)
next
  case IH: (Cons x xs)
  show ?case
  proof (cases maxOccurs)
    case (enat n)
    then show ?thesis using assms by (cases n; auto intro!: partial_function_mono IH)
  next
    case infinity with assms show ?thesis by (auto intro!: partial_function_mono IH)
  qed
qed

fun drop_xml_attributes :: "xml \<Rightarrow> xml" where
  "drop_xml_attributes (XML n atts cs) = (XML n [] cs)"
| "drop_xml_attributes x = x"

nonterminal xml_binds and xml_bind
syntax
  "_xml_block" :: "xml_binds \<Rightarrow> 'a" ("XMLdo {//(2  _)//}" [12] 62)
  "_xml_bind"  :: "pttrn \<Rightarrow> 'a \<Rightarrow> xml_bind" ("(2_ \<leftarrow>/ _)" 13)
  "_xml_let" :: "pttrn \<Rightarrow> 'a \<Rightarrow> xml_bind" ("(2let _ =/ _)" [1000, 13] 13)
  "_xml_final" :: "'a xmlst \<Rightarrow> xml_binds" ("_")
  "_xml_cons" :: "xml_bind \<Rightarrow> xml_binds \<Rightarrow> xml_binds" ("_;//_" [13, 12] 12)
  "_xml_begin" :: "tag \<Rightarrow> xml_binds \<Rightarrow> 'a" ("XMLbegin (_) {//(2 _)//}" [1000,12] 62)
  "_xml_enter" :: "tag \<Rightarrow> xml_binds \<Rightarrow> 'a" ("XMLenter (_) {//(2 _)//}" [1000,12] 62)

syntax (ASCII)
  "_xml_bind" :: "pttrn \<Rightarrow> 'a \<Rightarrow> xml_bind" ("(2_ <-/ _)" 13)

translations
  "_xml_block (_xml_cons (_xml_bind p x) (_xml_final e))"
    \<rightleftharpoons> "_xml_block (_xml_final (x (\<lambda>p. e)))"
  "_xml_block (_xml_cons (_xml_let p t) bs)"
    \<rightleftharpoons> "let p = t in _xml_block bs"
  "_xml_block (_xml_cons b (_xml_cons c cs))"
    \<rightleftharpoons> "_xml_block (_xml_cons b (_xml_final (_xml_block (_xml_cons c cs))))"
  "_xml_cons (_xml_let p t) (_xml_final s)"
    \<rightleftharpoons> "_xml_final (let p = t in s)"
  "_xml_block (_xml_final e)" \<rightharpoonup> "e"
  "_xml_begin t e" \<rightleftharpoons> "CONST xml_begin t (_xml_block e)"
  "_xml_enter t e" \<rightleftharpoons> "CONST xml_enter t (_xml_block e)"

end